(require "mapping")
(require "generic")
(provide "commit")

(defun check-commit-tag ()
  ;; assume we are in *mcvs-dir*
  (unless (probe-file *mcvs-version-name*)
    (error "Not in right dir."))
  (when (probe-file *mcvs-commit-tag-marker-name*)
    (error "Tagging with module version tag failed at the most recent
commit. Please see the retag-version command.")))

;(defun read-line-no-hang (stream)
;  (let ((line nil))
;    (do ((char (read-char stream nil) (read-char stream nil)))
;        ((or (null char) (char= #\Newline char))
;         (if (null char)
;             nil
;             (concatenate 'string (nreverse line))))
;      (push char line))))

(defun sandbox-status (stream)
  "Return the status of the sandbox. If there is nothing to commit
return :fresh. Else, if there is a newer version of the module return
:needs-update. Else return :needs-commit. It also detects changes to
the VERSION file and returns :corrupt."
  (let ((needs-commit nil)
        (needs-update nil))
    (do ((line (read-line stream nil) (read-line stream nil)))
        ((null line))
      (when (and (>= (length line) 5)
                 (string= (subseq line 0 5) "File:"))
        (cond ((search "Status: Needs Merge" line)
               (when (search "File: VERSION" line)
                 (return-from sandbox-status :corrupt))
               (setq needs-commit t
                     needs-update t))
              ((or (search "Status: Locally " line)
                   (search "Status: File had conflicts on merge" line))
               (when (search "File: VERSION" line)
                 (return-from sandbox-status :corrupt))
               (setq needs-commit t))
              ((search "Status: Needs" line)
               (setq needs-update t)))))
    (cond ((and needs-commit needs-update)
           :needs-update)
          (needs-commit
           :needs-commit)
          (t
           :fresh))))

(defun cvs-sandbox-status (cvs-options files)
  (with-input-from-program (s `("cvs" ,@(format-opt cvs-options)
                                "status" ,@files))
    (sandbox-status s)))

(defun mcvs-version-change-option (options)
  (find-bind (:test #'string= :key #'first)
	     (remainder (major "major") (minor "minor") (patch "patch"))
	     options
    (values
     (restart-case
      (cond
        ((> 2 (count nil (list major minor patch)))
         (error "More than one version change specified."))
        (major :major)
        (minor :minor)
        (patch :patch)
        (t (error "No version change specified.")))
      (assume-major () :report "Assume MAJOR change." :major)
      (assume-minor () :report "Assume MINOR change." :minor)
      (assume-patch () :report "Assume PATCH change." :patch))
     remainder)))

(defun tag-module-with-version (cvs-options version)
  ;; assume we are in *mcvs-dir*
  (do ()
      ((execute-program-xargs
        `("cvs" ,@(format-opt cvs-options)
          "tag" ,(version->tag version))))
    (restart-case
     (error "Tagging with module version tag for ~S failed." version)
     (retry () :report "Retry tagging."))))

(defun mcvs-commit-helper (files cvs-options command-options command-args)
  "If there were changes commit the module files and return the new
version, otherwise return nil. Version in *MCVS-VERSION-NAME* is
changed according to COMMAND-OPTIONS."
  ;; !!! write doc for error handling unwind/commit/tag ...
  (in-sandbox-root-dir
   (chdir *mcvs-dir*)
   (check-commit-tag)
   (case (cvs-sandbox-status cvs-options files)
     (:corrupt
      (error "VERSION file was locally modified."))
     (:fresh nil)
     (:needs-update
      (error "Up-to-date check failed! Can't commit, update first!"))
     (:needs-commit
      ;; Update the version file with the bumped version, do commit
      ;; and tag.
      (multiple-value-bind (version-change cvs-command-options)
          (mcvs-version-change-option command-options)
        (let* ((old-version (read-version-from-file *mcvs-version-name*))
               (new-version (ho-version:bump-version old-version
                                                     version-change)))
          (with-version-file (*mcvs-version-name* *mcvs-old-version-name*
                                                  :commit)
            (unless *dry-run-option*
              (write-version-to-file new-version *mcvs-version-name*))
            (chatter-debug "Invoking CVS.~%")
            (unless
                (execute-program-xargs
                 `("cvs" ,@(format-opt cvs-options)
                   "-e" "hoc-fi-editor"
                   "commit" ,@(format-opt cvs-command-options)
                   ,@command-args)
                 files)
              (error "Commit failed ...")))
          (unless *dry-run-option*
            (with-marker-file (*mcvs-commit-tag-marker-name*)
              (unwind-failure-protect
               (tag-module-with-version cvs-options new-version)
               (warn "~&Tagging with module version tag for ~S failed.
Without further action, it is impossible to reference this module version.
If the problem goes away it is possible to tag the module with the
'hoc retag-version' command in this sandbox if it still has the same
version.~%"
                     new-version))))))))))

(defun mcvs-commit (cvs-options command-options command-args)
  (multiple-value-bind (files meta-files)
      (prepare-file-list command-args
                         :global-if-empty-file-list t
                         :always-include-version-file t)
    (in-sandbox-root-dir
     (chatter-debug "Synchronizing.~%")
     (mapping-synchronize :filemap files
                          :direction :left)
     (mcvs-commit-helper (append (mapcar #'(lambda (x) 
                                             (basename (mapping-entry-id x)))
                                         files)
                                 meta-files)
                         cvs-options command-options nil)
     (chatter-debug "Synchronizing again.~%")
     (mapping-synchronize :filemap files
                          :direction :right))))

(defun mcvs-commit-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-commit cvs-options cvs-command-options mcvs-args))

(defun mcvs-retag-version (cvs-options command-options command-args)
  (declare (ignore command-options command-args))
  (in-sandbox-root-dir
   (chdir *mcvs-dir*)
   (unless (probe-file *mcvs-commit-tag-marker-name*)
     (restart-case
      (error "No marker file found. Tagging may not have failed.")
      (continue ()
                :report "Continue with tagging anyway?")))
   (tag-module-with-version cvs-options
                            (read-version-from-file *mcvs-version-name*))
   (delete-file *mcvs-commit-tag-marker-name*)))

(defun mcvs-retag-version-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-retag-version cvs-options cvs-command-options mcvs-args))

(defconstant *commit-help*
  "Syntax:

  hoc commit [ options ] [ file list ]

Options:

  -m message       Log message.
  -F log-file      Read the log message from file.
  --major          Bump the major version.
  --minor          Bump the minor version.
  --patch          Bump the patch level.

Semantics:

  The commit command updates the repository with the contents of the
  sandbox and changes the module version according to the options.

  If some files are explicitly specified, the meta files are not
  committed by default. Use --meta counter this.

  After the commit is done the module is tagged to allow retrieval
  by module version. The tagging operation may fail even if commit
  succeeded. In this case the retag-version command may help.")

(defconstant *retag-version-help*
  "Syntax:

  hoc retag-version

Options:

Semantics:

  If the tagging operation after a commit fails the committed version
  of the module cannot be retrieved by module version (e.g. hoc up -r
  '#v1.2.3'). If since the failed commit no further commits or updates
  have taken place in the same sandbox, running retag-version creates
  the same tag commit should have created.")
