;;; This source file is part of the Meta-CVS program, 
;;; which is distributed under the GNU license.
;;; Copyright 2002 Kaz Kylheku

(require "dirwalk")
(require "chatter")
(require "mapping")
(require "options")
(require "generic")
(provide "update")

(defun mcvs-update-helper (cvs-options cvs-update-options files meta-files)
  (super-restart-case
   (progn
     (chatter-debug "Invoking CVS.~%")
     (unless (execute-program-xargs
              `("cvs" ,@(format-opt cvs-options)
                "up" ,@(format-opt cvs-update-options))
              `(,@meta-files
                ,@(mapcar #'(lambda (x)
                              (basename (mapping-entry-id x)))
                          files)))
       (error "CVS update failed.")))
   (continue () 
             :report
             "Update file structure and re-synchronize." 
             (unwind))
   (retry () 
          :report "Try invoking CVS again." 
          (retry))))

(defun sticky-to-non-branch? ()
  (if-bind (sticker (what-are-we-sticky-to))
           (if (eq :branch (first sticker))
               nil
               t)
           nil))

(defun updating-update-options? (update-options)
  "See if there are options among UPDATE-OPTIONS that change file
versions: -A -D -r."
  (find-if #'(lambda (x)
               (member x (list "A" "D" "r") :test #'string=))
           update-options :key #'first))

(defun mcvs-update (&optional cvs-options cvs-update-options file-args)
  (if (find "p" cvs-update-options :key #'first :test #'string=)
      (mcvs-generic "update" cvs-options cvs-update-options nil file-args)
      (let ((merging? (find "j" cvs-update-options
                            :key #'first :test #'string=)))
        (multiple-value-bind (files meta-files global?)
            (prepare-file-list
             file-args
             :global-if-empty-file-list t
             :always-include-version-file t)
          (in-sandbox-root-dir
           ;; Push changes in tree to CVS sandbox, so they can be merged
           ;; with stuff coming from repository.
           (chatter-debug "Synchronizing.~%")
           (mapping-synchronize :filemap files
                                :direction :left)
           (current-dir-restore 
            (chdir *mcvs-dir*)
            (check-commit-tag)
            (if (not merging?)
                (progn
                  (when (not global?)
                    (error "Refusing to update only some of the files."))
                  (with-version-file (*mcvs-version-name*
                                      *mcvs-old-version-name*
                                      :update
                                      :restore-on-failure nil)
                    (mcvs-update-helper cvs-options cvs-update-options
                                        files meta-files)))
                (progn
                  ;; Cvs does an update for the files being merged if
                  ;; we are not careful which would lead to a sandbox
                  ;; with mixed file versions. So add -rTHISVERSION to
                  ;; the command and unstick afterwards.
                  (when (sticky-to-non-branch?)
                    ;; if cvs merges a file addition to a sticky sandbox it gets
                    ;; a strange branch version
                    (error "Refusing to merge while being sticky due to CVS peculiarities.
See 'hoc unstick'"))
                  (when (updating-update-options? cvs-update-options)
                    (error "Do not mix updating and merging."))
                  (unwind-protect-file *mcvs-version-name*
                                       *mcvs-old-version-name*
                    (let ((version-tag (version->tag (read-version-from-file
                                                      *mcvs-version-name*))))
                      (mcvs-update-helper cvs-options
                                        `(("r" ,version-tag)
                                          ,@cvs-update-options)
                                        files meta-files)))
                  (unstick))))
           (chatter-debug "Updating file structure.~%")
           (mapping-update)
           (chatter-debug "Synchronizing again.~%")
           (mapping-synchronize :filemap files
                                :direction :right))))))

(defun mcvs-update-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-update cvs-options (subst-version-with-tag cvs-command-options)
               mcvs-args))
