;;;; general purpose utilities

(provide "util")

(defmacro lastone (list)
  "Return the last element of LIST."
  `(first (last ,list)))

(defmacro if-bind ((var value) then &optional else)
  `(let ((,var ,value))
    (if ,var
        ,then
        ,else)))

(defmacro when-bind ((var value) &body body)
  `(let ((,var ,value))
    (when ,var
      ,@body)))

(defmacro when-bind* (bindings &body body)
  "Bind BINDINGS one by one until the first NIL value is
encountered. If none of them is NIL, then eval BODY in this context."
  (if (null bindings)
      `(progn ,@body)
      `(when-bind ,(car bindings) (when-bind ,(cdr bindings) ,@body))))

(defun subst-fn (pred fn tree) 
  (cond ((funcall pred tree)
         (funcall fn tree))
        ((atom tree) tree)
        (t (let ((a (funcall #'subst-fn pred fn (car tree)))
                 (d (funcall #'subst-fn pred fn (cdr tree))))
             (if (and (eql a (car tree))
                      (eql d (cdr tree)))
                 tree
                 (cons a d))))))

(defun pathname-file-p (pathname)
  "See if PATHNAME denotes a file (as opposed to a directory) without
doing any I/O."
  (or (pathname-name pathname) (pathname-type pathname)))

(defun pathname-directory-p (pathname)
  "See if PATHNAME denotes a directory without doing any I/O."
  (not (pathname-file-p pathname)))

(defun pathname-as-file (pathname)
  "If there is no name and type in PATHNAME, then interpret the last directory
  as the filename."
  (if (and (zerop (length (file-namestring pathname)))
           (> (length (pathname-directory pathname)) 1))
      (make-pathname :host (pathname-host pathname)
                     :device (pathname-device pathname)
                     :directory (butlast (pathname-directory pathname))
                     :name (first (last (pathname-directory pathname)))
                     :type nil
                     :version nil)
      pathname))

;;; by Erik Naggum from comp.lang.lisp
(defun pathname-as-directory (pathname)
  "Interpret pathname as a directory even without trailing slahes and such."
  (setq pathname (pathname pathname))
  (let ((filename (file-namestring pathname)))
    (if (and filename (> (length filename) 0))
        (make-pathname :defaults pathname
                       :directory (append (pathname-directory pathname)
                                          (list filename))
                       :name nil :type nil :version nil)
        pathname)))

(defun add-type-to-pathname (pathname type)
  "Add TYPE to PATHNAME by setting pathname-type to TYPE and
concatenating the old type to pathname-name."
  (make-pathname :name (if (pathname-type pathname)
                           (concatenate 'string
                                        (pathname-name pathname)
                                        "."
                                        (pathname-type pathname))
                           (pathname-name pathname))
                 :type type
                 :defaults pathname))

(defmacro unwind-failure-protect (protected-from cleanup-form)
  "Like unwind-protect, but CLEANUP-FORM is evaluated only when
PROTECTED-FROM fails (i.e. does a non-local exit)."
  (let ((success (gensym))
        (r (gensym)))
    `(let ((,success nil)
           (,r nil))
      (unwind-protect
           (progn
             (setq ,r (multiple-value-list ,protected-from))
             (setq ,success t)
             (values-list ,r))
        (when (not ,success)
          ,cleanup-form)))))

(defun touch-file (pathname)
  (with-open-file (s pathname :direction :output :if-exists :append
                     :if-does-not-exist :create)))

(defun copy-file (from to)
  (when (not (execute-program `("cp" ,(namestring from) ,(namestring to))))
    (error "Copying failed.")))

(defun rename-file2 (from to)
  (when (not (execute-program `("mv" ,(namestring from) ,(namestring to))))
    (error "Renaming failed.")))

(defmacro unwind-protect-file (pathname backup-pathname &body body)
  ;; !!! gensyms
  `(progn
    (delete-file ,backup-pathname)
    (rename-file2 ,pathname ,backup-pathname)
    (unwind-protect
         (progn
           (copy-file ,backup-pathname ,pathname)
           ,@body)
      (progn
        (delete-file ,pathname)
        (rename-file2 ,backup-pathname ,pathname)))))

(defmacro unwind-failure-protect-file (pathname backup-pathname &body body)
  ;; !!! gensyms
  `(progn
    (delete-file ,backup-pathname)
    (rename-file2 ,pathname ,backup-pathname)
    (unwind-failure-protect
         (progn
           (copy-file ,backup-pathname ,pathname)
           ,@body)
      (progn
        (delete-file ,pathname)
        (rename-file2 ,backup-pathname ,pathname)))))

(defmacro with-marker-file ((pathname) &body forms)
  "Create a file by the PATHNAME that marks the succesful completion
of FORMS."
  `(prog2
    (touch-file ,pathname)
    (progn ,@forms)
    (delete-file ,pathname)))

(defmacro add-hook (hook fn)
  "Add FN to HOOK."
  `(setq ,hook (adjoin ,fn ,hook)))

(defun remove-hook (hook fn)
  "Remove FN from HOOK."
  `(setq ,hook (remove ,fn ,hook)))

(defun run-hook (hook &rest args)
  "Call each function on HOOK with ARGS."
  (let ((results ()))
    (dolist (fn hook (nreverse results))
      (with-simple-restart (continue "Continue running this hook.")
        (push (apply fn args) results)))))
                    
