;;;; utilities for dealing with versions

(provide "version-util")

(defvar *version-change-hook* ()
  "Functions on this hook are called when a module's version
changes. The argument list is (module-name old-version
new-version). This one is only kept around for backward
compatibility.")

(defvar *operation-hook* ()
  "Functions on this hook are called when a certain operation takes
place. The argument list is (module-name old-version new-version
operation) where operation is one of :checkout :update
:commit. Old-version is nil on checkout.")

(defvar *version-changes* ())
(defvar *operations* ())

(defmacro with-version-file ((file backup-file operation
                                   &key (restore-on-failure t))
                             &body body)
  "Restore the version FILE in case of failure of BODY (if
RESTORE-ON-FAILURE), otherwise push version change info on
*version-changes* and return the new and old versions."
  ;; !!! order of evaluation
  (assert (member operation '(:checkout :update :commit)))
  (let ((new-version (gensym))
        (version-file (gensym))
        (old-version (gensym))
        (old-version-file (gensym)))
    `(let* ((,version-file ,file)
            (,old-version (read-version-from-file ,version-file))
            (,old-version-file ,backup-file))
      (,@(if restore-on-failure
             `(unwind-failure-protect-file ,version-file ,old-version-file)
             '(progn))
       ,@body)
      (let ((,new-version (read-version-from-file ,version-file)))
        (unless (ho-version:version= ,old-version ,new-version)
          (push (list (module-name) ,old-version ,new-version)
                *version-changes*))
        (push (list (module-name) ,old-version ,new-version ,operation)
              *operations*)
        (values ,new-version ,old-version)))))

(defun run-version-change-hook ()
  (handler-bind ((error #'(lambda (condition)
                            (format *error-output*
                                    "Hook failed: ~A~%" condition)
                            (continue))))
    (dolist (vc *version-changes*)
      (apply #'run-hook *version-change-hook* vc))))

(defun run-operation-hook ()
  (handler-bind ((error #'(lambda (condition)
                            (format *error-output*
                                    "Hook failed: ~A~%" condition)
                            (continue))))
    (dolist (o *operations*)
      (apply #'run-hook *operation-hook* o))))

(defun write-version-to-file (version pathname)
  "Write VERSION to PATHNAME."
  (with-open-file (stream pathname :direction :output :if-exists :supersede)
    (format stream "; NEVER edit this file by hand~%")
    (prin1 version stream)))

(defun read-version-from-file (pathname)
  "Read version from PATHNAME. Return a second value indicating if the
file has cvs conflict."
  (labels ((check-token (token expected-token)
             (unless (eql token expected-token)
               (error "Unexpected token ~S in VERSION file." token)))
           (expect-token (stream expected-token)
             (check-token (read stream) expected-token)))
    (with-open-file (s pathname :direction :input)
      (let ((token (read s)))
        (cond ((typep token 'ho-version:version)
               (values token nil))
              (t
               (check-token token '<<<<<<<)
               (expect-token s 'VERSION)
               (values (read s) t)))))))

(defun version->tag (version)
  "Convert VERSION to a string suitable for cvs tag as a parameter."
  (let ((sv (lastone (ho-version:version-part-list version))))
    (apply #'format nil "~[~A~;MAIN~]-~D-~D-~D!"
           (append
            (if (ho-version:simple-version-variant sv)
                (list 0 (ho-version:simple-version-variant sv))
                (list 1))
            (list (ho-version:simple-version-major sv)
                  (ho-version:simple-version-minor sv)
                  (ho-version:simple-version-patch sv))))))

(defun make-pred-nth (pred n)
  "Create a predicate that's true for the Nth time PRED is true."
  #'(lambda (x)
      (when (funcall pred x)
        (decf n)
        (zerop n ))))

(defun split-version-tag (tag)
  "Split tag at the end branch name: the 3rd '-' from the end."
  (let ((end-of-branch-name
         (position-if (make-pred-nth #'(lambda (c) (char= c #\-)) 3)
                      tag
                      :from-end t)))
    (list (subseq tag 0 end-of-branch-name)
          (mapcar #'read-from-string
                  (split-fields (subseq tag
                                        (1+ end-of-branch-name)
                                        (1- (length tag)))
                                "-")))))

(defun tag->version (tag)
  "Convert TAG to a version or return nil."
  (ignore-errors
    (unless (char= (elt tag (1- (length tag))) #\!)
      (error "Malformed version tag ~S" tag))
    (destructuring-bind (branch-name versions) (split-version-tag tag)
      (unless (= 3 (length versions))
        (error "Malformed version tag ~S" tag))
      (ho-version:make-version
       :part-list
       (list (ho-version:make-simple-version
              :variant (if (string= branch-name "MAIN")
                           nil
                           branch-name)
              :major (first versions)
              :minor (second versions)
              :patch (third versions)))))))

(defstruct version-tree-node
  version
  cvs-version
  successor
  branches)

(defun find-versions-in-tree (pred tree)
  "Returns a subtree of this tree (possibly nil) and a forest (list of trees)."
  (if tree
      (multiple-value-bind (successor-subtree successor-forest)
          (find-versions-in-tree pred (version-tree-node-successor tree))
        (let ((branch-subtrees ())
              (forest successor-forest))
          (dolist (b (version-tree-node-branches tree))
            (multiple-value-bind (s f)
                (find-versions-in-tree pred b)
              (setq branch-subtrees
                    (nconc branch-subtrees (if s (list s) s)))
              (setq forest (nconc forest f))))
          (if (funcall pred tree)
              (values (make-version-tree-node
                       :version (version-tree-node-version tree)
                       :cvs-version (version-tree-node-cvs-version tree)
                       :successor successor-subtree
                       :branches branch-subtrees)
                      forest)
              (values nil
                      (nconc forest
                             (if successor-subtree
                                 (list successor-subtree)
                                 nil)
                             branch-subtrees)))))
      (values nil nil)))

(defun version-tree-ranges (tree)
  "Return a list of single nodes and (first-node-on-branch
. last-node-on-branch) pairs that cover the whole tree."
  (labels ((cvs-version-same-branch (v1 v2)
             (string=
              (subseq v1 0 (position #\. v1 :from-end t))
              (subseq v2 0 (position #\. v2 :from-end t))))
           (create-range (start end)
             (if (eq start end)
                 start
                 (cons start end)))
           (xxx (tree branch-start)
             (let ((branch-ranges (mapcan #'version-tree-ranges
                                          (version-tree-node-branches tree))))
               (if (not (version-tree-node-successor tree))
                   (cons (create-range branch-start tree)
                         branch-ranges)
                   ;; Because of the !@#$ vendor branches #v0.0.0 is
                   ;; 1.1.1.1 and #v0.0.1 is 1.2, so let's create a
                   ;; new range if the cvs branch changes midbranch.
                   ;; Well, this would work for the VERSION file,
                   ;; because it always changes, but for other files
                   ;; any version on the trunk might still be
                   ;; 1.1.1.1. So, let's list them all one by one.
                   (if (ho-version:version-variant
                        (version-tree-node-version tree))
;;                             (cvs-version-same-branch
;;                              (version-tree-node-cvs-version
;;                               (version-tree-node-successor tree))
;;                              (version-tree-node-cvs-version branch-start)))
                       (nconc (xxx (version-tree-node-successor tree)
                                   branch-start)
                              branch-ranges)
                       (nconc (list (create-range branch-start tree))
                              (xxx (version-tree-node-successor tree)
                                   (version-tree-node-successor tree))
                              branch-ranges))))))
    (xxx tree tree)))

(defun subst-version-with-tag (options)
  "Substitute option value strings (that are really #v versions)
following -r options with the appropriate tag names."
  (labels ((pred (x) (and (stringp x)
                          (char= (char x 0) #\#)
                          (char= (char x 1) #\v)))
           (trans (x) (version->tag (read-from-string x))))
    (subst-fn #'(lambda (x) (and (listp x)
                                 (and (stringp (first x)))
                                 (or (string= (first x) "r")
                                     (string= (first x) "j"))))
              #'(lambda (x) `(,(first x) ,@(subst-fn #'pred #'trans (rest x))))
              options)))
