(defpackage "HO-VERSION"
  (:use "COMMON-LISP")
  (:export "SIMPLE-VERSION"
           "MAKE-SIMPLE-VERSION"
           "SIMPLE-VERSION-VARIANT"
           "SIMPLE-VERSION-MAJOR"
           "SIMPLE-VERSION-MINOR"
           "SIMPLE-VERSION-PATCH"
           "VALID-SIMPLE-VERSION-VARIANT-STRING?"
           "VERSION"
           "MAKE-VERSION"
           "SIMPLIFY-VERSION"
           "VERSION-VARIANT"
           "VERSION-PART-LIST"
           "CANONICAL-VERSION?"
           "BUMP-VERSION"
           "BRANCH-VERSION"
           "VERSION="
           "VERSION>="))

(in-package "HO-VERSION")

(defmacro lastone (list)
  "Return the last element of LIST."
  `(first (last ,list)))

;;; from comp.lang.lisp
;;; delimiter can be a predicate
(defun split-sequence (delimiter seq
                                 &key
                                 (empty-marker nil keep-empty-subseqs)
                                 (from-end nil)
                                 (start 0)
                                 (end nil)
                                 (test nil test-supplied)
                                 (test-not nil test-not-supplied)
                                 (key nil key-supplied)
                                 &aux
                                 (len (length seq)))
  "Return list of subsequences in SEQ delimited by DELIMITER (which
can either be an element or a predicate). If an EMPTY-MARKER is
supplied, empty subsequences will be represented by EMPTY-MARKER,
otherwise they will be discarded.  All other keywords work analogously
to POSITION."
  ;;(declare (optimize (speed 3) (safety 0) (space 0) (debug 0)))
  (unless end (setq end len))
  (when from-end
    (setf seq (reverse seq))
    (psetf start (- len end)
           end (- len start)))
  (loop with other-keys = (nconc (when test-supplied (list :test test))
                                 (when test-not-supplied
                                   (list :test-not test-not))
                                 (when key-supplied (list :key key)))
        for left = start then (+ right 1)
        for right = (min (or (apply (if (functionp delimiter)
                                        #'position-if
                                        #'position)
                                    delimiter seq
                                    :start left other-keys)
                             len)
                         end)
        if (< left right)
        collect (let ((subseq (subseq seq left right)))
                  (if from-end
                      (nreverse subseq)
                      subseq))
        else when keep-empty-subseqs collect empty-marker
        until (eq right end)))

(defun char->string (char)
  "Create a string of a single CHAR."
  (make-string 1 :initial-element char))

(defun read-chars (pred &optional (stream nil) (eof-error-p t)
                        (eof-value nil) (recursive-p nil))
  "Read characters defined by PRED from STREAM and return them as a string."
  (labels ((rc () (read-char stream eof-error-p eof-value recursive-p))
           (peek-pred ()
             (let ((c (peek-char nil stream eof-value eof-value recursive-p)))
               (if c
                   (funcall pred c)
                   nil))))
    (if (peek-pred)
        (do ((s (char->string (rc))
                (concatenate 'string s (char->string (rc)))))
            ((not (peek-pred)) s))
        "")))

;;;; simple version
(defun print-simple-version (sv stream)
  "Print SV to STREAM in the form of [VARIANT:]MAJOR.MINOR.PATCH."
  (princ (apply #'format nil "~[~A:~;~]~D.~D.~D"
                (append
                 (if (simple-version-variant sv)
                     (list 0 (simple-version-variant sv))
                     (list 1))
                 (list (simple-version-major sv)
                       (simple-version-minor sv)
                       (simple-version-patch sv))))
         stream))

(defun valid-simple-version-variant-char? (char)
  "See if CHAR is valid in variants."
  (or (find char "-_")
      (alpha-char-p char)
      (digit-char-p char)))
  
(defun valid-simple-version-variant-string? (variant)
  (every #'valid-simple-version-variant-char? variant))

(deftype simple-version-variant ()
  `(or null (satisfies valid-simple-version-variant-string?)))

(defstruct (simple-version
             (:constructor)
             (:constructor make-simple-version-boa (variant major minor
                                                            patch)))
  "A simple version is a tuple of 3 numbers plus a variant:
- the major (first) version number is bumped for incompatible changes
  is introduced
- the minor (second) version number is bumped for upward compatible changes
- the patch (third) version number is bumped for compatible changes
Whenever a number in a simple version changes all subsequent numbers
are reset to 0."
  (variant nil :type simple-version-variant)
  (major 0 :type integer)
  (minor 0 :type integer)
  (patch 0 :type integer))

(defun simple-version= (sv1 sv2)
  "See if SV1 is equal to SV2."
  (and (or (and (not (simple-version-variant sv1))
                (not (simple-version-variant sv2)))
           (string= (simple-version-variant sv1)
                    (simple-version-variant sv2)))
       (= (simple-version-major sv1) (simple-version-major sv2))
       (= (simple-version-minor sv1) (simple-version-minor sv2))
       (= (simple-version-patch sv1) (simple-version-patch sv2))))

(defun simple-version>= (sv1 sv2 &optional (difference :major))
  "See if SV1 is upward compatible with SV2 with regards to
DIFFERENCE. Differences up from DIFFERENCE are considered to be
incompatible, otherwise lexicographical comparison is performed."
  (let ((equal-until (position difference '(:major :minor :patch)))
        (>=? :unkown))
    (assert equal-until)
    (if (not (equal (simple-version-variant sv1)
                    (simple-version-variant sv2)))
        nil
        (labels ((simple-version->list (sv)
                   (list (simple-version-variant sv)
                         (simple-version-major sv)
                         (simple-version-minor sv)
                         (simple-version-patch sv))))
          (do ((l1 (cdr (simple-version->list sv1)) (cdr l1))
               (l2 (cdr (simple-version->list sv2)) (cdr l2))
               (i 0 (1+ i)))
              ((or (endp l1) (not (eq >=? :unkown)))
               (if (not >=?) nil t))
            (setq >=?
                  (if (< i equal-until)
                      (if (not (= (car l1) (car l2)))
                          nil
                          :unkown)
                      (cond ((> (car l1) (car l2))
                             t)
                            ((= (car l1) (car l2))
                             :unkown)
                            (t
                             nil)))))))))

(defun bump-simple-version-major (sv)
  "Bump the major version, reset minor version and patch level"
  (make-simple-version :variant (simple-version-variant sv)
                       :major (1+ (simple-version-major sv))
                       :minor 0
                       :patch 0))

(defun bump-simple-version-minor (sv)
  "Bump the minor version, reset patch level"
  (make-simple-version :variant (simple-version-variant sv)
                       :major (simple-version-major sv)
                       :minor (1+ (simple-version-minor sv))
                       :patch 0))

(defun bump-simple-version-patch (sv)
  "Bump the patch level"
  (make-simple-version :variant (simple-version-variant sv)
                       :major (simple-version-major sv)
                       :minor (simple-version-minor sv)
                       :patch (1+ (simple-version-patch sv))))

(defun bump-simple-version (sv part)
  "Bump the version number in SV specified by PART.  Valid values for
PART are :major :minor and :patch."
  (funcall (case part
             (:major #'bump-simple-version-major)
             (:minor #'bump-simple-version-minor)
             (:patch #'bump-simple-version-patch)
             (t (assert nil () "Invalid version part: ~S" part)))
           sv))

(defun change-simple-version-variant (sv variant)
  "Return a new simple version with changed variant."
  (make-simple-version :variant variant
                       :major (simple-version-major sv)
                       :minor (simple-version-minor sv)
                       :patch (simple-version-patch sv)))

;;;; version
(defun print-version (version stream)
  "Print VERSION to STREAM in the form of #vSV(1)/.../SV(n)."
  (princ "#v" stream)
  (let ((first t))
    (dolist (e (version-part-list version))
      (if first
          (setq first nil)
          (princ "/" stream))
      (print-simple-version e stream))))

(defun valid-version-part-list? (part-list)
  "See if there are anomalies: zero parts, no variant at positions
other than the first, variant at the first positions when there are
more than one parts."
  (and (>= (length part-list) 1)
       (notany #'(lambda (sv) (null (simple-version-variant sv)))
               (rest part-list))))

(deftype version-part-list ()
  `(and list (satisfies valid-version-part-list?)))

(defstruct (version (:print-object print-version))
  "A version is a list of simple versions.  All but the last simple
version denote branching points (the first being the branching point
from the trunk).  The last simple version determines the position on
the branch."
  (part-list (list (make-simple-version)) :type version-part-list))

(defun version-variant (version)
  "Return the variant of last simple version of VERSION."
  (simple-version-variant
    (first (last (version-part-list version)))))

(defun simplify-version (version)
  "Return a new version that contains only the last simple version of
VERSION."
  (make-version :part-list (list
                            (copy-simple-version
                             (lastone (version-part-list version))))))

(defun canonical-version? (version)
  "A version is canonical if it contains the all its ancestors in the
version tree. A non-canonical versions are easy to detect: its first
simple version component has a NIL variant.

Examples of canonical versions:
  #v0.0.0/X:5.6.7/Y:3.2.1
  #v0.0.0
Examples of non-canonical versions:
  #vX:5.6.7
  #vY:3.2.1"
  (null (simple-version-variant (first (version-part-list version)))))

(defun bump-version (version part)
  "Bump the version number in VERSION specified by PART.  Valid values
for PART are :major :minor and :patch."
  (make-version
   :part-list
   (concatenate 'list
                (butlast (version-part-list version))
                (list (bump-simple-version
                       (first (last (version-part-list version)))
                       part)))))

(defun branch-version (version branch-name)
  "Create a new branch from VERSION by appending a fresh
simple-version to it."
  (if (canonical-version? version)
      (make-version :part-list
                    (concatenate 'list
                                 (version-part-list version)
                                 (list (make-simple-version :variant
                                                            branch-name))))
      (error "Cannot branch non-canonical version")))

(defun version-char-p (char)
  "See if CHAR is a valid character in versions."
  (or (valid-simple-version-variant-char? char)
      (find char "./:")
      (alpha-char-p char)))

(defun parse-simple-version (string)
  "Parse STRING of the form [V:]A.B.C into o SIMPLE-VERSION."
  (let* ((variant-end-pos (position #\: string))
         (version-segments (split-sequence #\. string
                                           :start (if variant-end-pos
                                                      (1+ variant-end-pos)
                                                      0))))
    ;; !!! parse errors
    (handler-case
        (make-simple-version :variant (if variant-end-pos
                                          (subseq string 0 variant-end-pos)
                                          nil)
                             :major (parse-integer (first version-segments))
                             :minor (parse-integer (second version-segments))
                             :patch (parse-integer (third version-segments)))
      (error () (error (format nil "`#v~A' is not a valid version."
                               string))))))

(defun parse-version (string)
  "Parse STRING of the form [V(1):]A(1).B(1).C(1)/.../[V(n):]A(n).B(n).C(n)
  into a version."
  (let ((v (make-version
            :part-list
            (map 'list #'parse-simple-version (split-sequence #\/ string)))))
    (assert (valid-version-part-list? (version-part-list v)))
    v))

(defun |#v-reader| (stream subchar arg)
  "Read a version from STREAM."
  (declare (ignore subchar arg))
  (parse-version (read-chars #'version-char-p stream t nil t)))

;;; hook the version reader up
(set-dispatch-macro-character #\# #\v #'|#v-reader|)

(defun version= (version1 version2)
  "See if VERSION1 is equal to VERSION2. Canonical versions are
compared simple version by simple version. If any of the two versions
is non-canonical, then only the last simple versions are compared.
For example #v0.0.0/X:1.2.3 #vX:1.2.3 are equal."
  (labels ((part-list= (pl1 pl2)
             (cond
               ((and (endp pl1) (endp pl2))
                t)
               ((and pl1 pl2)
                (and (simple-version= (first pl1) (first pl2))
                     (part-list= (rest pl1) (rest pl2))))
               (t nil))))
    (cond
      ((and (canonical-version? version1)
            (canonical-version? version2))
       (part-list= (version-part-list version1) (version-part-list version2)))
      (t
       (simple-version= (lastone (version-part-list version1))
                        (lastone (version-part-list version2)))))))

(defun version>= (version1 version2 &optional (difference :major))
  "See if VERSION1 is upward compatible with VERSION2 with regards to
DIFFERENCE. This means that on the version tree VERSION1 is a
descendant of VERSION2 and there are no more significant changes than
DIFFERENCE between them.

A = (A(1) A(2) ... A(n))
B = (B(1) B(2) ... B(m))
A >= B, iff
  n >= m,
  for all i in [1, m - 1]: A(i) = B(i),
  A(m) >= B(m),
  for all i in [m + 1, n]: A(i) >= (Av(i) 0 0 0),
  where Av(i) is the variant in A(i).

The above is for canonical versions. In general the comparsion is
performed on the common subtree of the two versions. This means that
some elements from the beginning of either A or B are removed until
variants at the beginning are the same."
  (labels ((sv>= (sv1 sv2)
             (simple-version>= sv1 sv2 difference))
           (version-part-list>= (v1 v2)
             (cond ((and (endp v1) (endp v2)) t)
                   ((and (not (endp v1)) (not (endp v2)))
                    (and (if (rest v2)
                             (simple-version= (first v1) (first v2))
                             (sv>= (first v1) (first v2)))
                         (version-part-list>= (rest v1) (rest v2))))
                   ((endp v2)
                    (and (sv>= (change-simple-version-variant (first v1) nil)
                               (make-simple-version))
                         (version-part-list>= (rest v1) nil)))
                   (t nil))))
    (let ((l1 (version-part-list version1))
          (l2 (version-part-list version2)))
      (let ((l1-tail (member (simple-version-variant (first l2))
                             l1
                             :key #'simple-version-variant
                             :test #'equal))
            (l2-tail (member (simple-version-variant (first l1))
                             l2
                             :key #'simple-version-variant
                             :test #'equal)))
        (cond (l1-tail
               (version-part-list>= l1-tail l2))
              (l2-tail
               (version-part-list>= l1 l2-tail))
              ((canonical-version? version1)
               nil)
              (t
               (error "Cannot compare version ~S to version ~S."
                      version1 version2)))))))
